#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import polyaxon_sdk


class V1ArtifactKind(polyaxon_sdk.V1ArtifactKind):
    @classmethod
    def is_single_file_event(cls, kind: str) -> bool:
        return kind in {
            V1ArtifactKind.HTML,
            V1ArtifactKind.TEXT,
            V1ArtifactKind.HISTOGRAM,
            V1ArtifactKind.CHART,
            V1ArtifactKind.CONFUSION,
            V1ArtifactKind.CURVE,
            V1ArtifactKind.METRIC,
            V1ArtifactKind.SYSTEM,
        }

    @classmethod
    def is_single_or_multi_file_event(cls, kind: str) -> bool:
        return kind in {
            V1ArtifactKind.MODEL,
            V1ArtifactKind.DATAFRAME,
            V1ArtifactKind.AUDIO,
            V1ArtifactKind.VIDEO,
            V1ArtifactKind.IMAGE,
            V1ArtifactKind.CSV,
            V1ArtifactKind.TSV,
            V1ArtifactKind.PSV,
            V1ArtifactKind.SSV,
        }

    @classmethod
    def is_dir(cls, kind: str) -> bool:
        return kind in {
            V1ArtifactKind.TENSORBOARD,
            V1ArtifactKind.DIR,
        }

    @classmethod
    def is_file(cls, kind: str) -> bool:
        return kind in {
            V1ArtifactKind.DOCKERFILE,
            V1ArtifactKind.FILE,
            V1ArtifactKind.ENV,
        }

    @classmethod
    def is_file_or_dir(cls, kind: str) -> bool:
        return kind in {
            V1ArtifactKind.DATA,
            V1ArtifactKind.MODEL,
        }



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple, Union

from traceml.processors.errors import NUMPY_ERROR_MESSAGE, PANDAS_ERROR_MESSAGE
from traceml.processors.units_processors import to_percentage

try:
    import numpy as np
except ImportError as e:
    raise ImportError(NUMPY_ERROR_MESSAGE) from e

try:
    import pandas as pd

    from pandas.api import types as pd_types
except ImportError as e:
    raise ImportError(PANDAS_ERROR_MESSAGE) from e

DF_TYPE_BOOL = "bool"
DF_TYPE_NUMERIC = "numeric"
DF_TYPE_DATE = "date"
DF_TYPE_CATEGORICAL = "categorical"
DF_TYPE_CONSTANT = "constant"
DF_TYPE_UNIQUE = "unique"

EXCLUDE = "exclude"
INCLUDE = "include"
RAISE = "raise"
ALL = "all"


def df_has_column(df: pd.DataFrame, column: Union[int, str]):
    if not isinstance(column, (int, str)):
        raise ValueError("{} is not a valid column".format(column))
    return column in df.columns


def get_df_columns(
    df: pd.DataFrame,
    usage: Union[INCLUDE, EXCLUDE, ALL],
    columns: Union[pd.Index, List[str]] = None,
):
    """
    Returns a `pd.DataFrame.columns`.

    Args:
        df: dataframe to select columns from
        usage: should be a value from [ALL, INCLUDE, EXCLUDE].
             this value only makes sense if attr `columns` is also set.
             otherwise, should be used with default value ALL.
        columns:
             * if `usage` is all, this value is not used.
             * if `usage` is INCLUDE, the `df` is restricted to the intersection
             between `columns` and the `df.columns`
             * if usage is EXCLUDE, returns the `df.columns` excluding these `columns`
    Returns:
        `pd.DataFrame` columns, excluding `target_column` and `id_column` if given.
        `pd.DataFrame` columns, including/excluding the `columns` depending on `usage`.
    """
    columns_excluded = pd.Index([])
    columns_included = df.columns

    if usage == INCLUDE:
        try:
            columns_included = columns_included.intersection(pd.Index(columns))
        except TypeError:
            pass
    elif usage == EXCLUDE:
        try:
            columns_excluded = columns_excluded.union(pd.Index(columns))
        except TypeError:
            pass

    columns_included = columns_included.difference(columns_excluded)
    return columns_included.intersection(df.columns)


def get_df_uniques(df: pd.DataFrame):
    return pd.Series(dict((c, df[c].nunique()) for c in df.columns), name="uniques")


def get_df_missing(df: pd.DataFrame, df_length: int = None, df_counts: int = None):
    if df_length is None:
        df_length = len(df)
    if df_counts is None:
        df_counts = df.count()
        df_counts.name = "counts"
    count = df_length - df_counts
    count.name = "missing"
    perc = (count / df_length).apply(to_percentage)
    perc.name = "missing_perc"
    return pd.concat([count, perc], axis=1, sort=True)


def get_df_column_stats(df: pd.DataFrame, df_counts: int = None) -> pd.DataFrame:
    def get_df_columns_info():
        column_info = {}
        column_info[DF_TYPE_CONSTANT] = stats["uniques"][stats["uniques"] == 1].index
        column_info[DF_TYPE_BOOL] = stats["uniques"][stats["uniques"] == 2].index
        rest_columns = get_df_columns(
            df, EXCLUDE, column_info["constant"].union(column_info["bool"])
        )
        column_info[DF_TYPE_NUMERIC] = pd.Index(
            [c for c in rest_columns if pd_types.is_numeric_dtype(df[c])]
        )
        rest_columns = get_df_columns(df[rest_columns], EXCLUDE, column_info["numeric"])
        column_info[DF_TYPE_DATE] = pd.Index(
            [c for c in rest_columns if pd_types.is_datetime64_dtype(df[c])]
        )
        rest_columns = get_df_columns(df[rest_columns], EXCLUDE, column_info["date"])
        unique_columns = stats["uniques"][rest_columns] == stats["counts"][rest_columns]
        column_info[DF_TYPE_UNIQUE] = stats["uniques"][rest_columns][
            unique_columns
        ].index
        column_info[DF_TYPE_CATEGORICAL] = stats["uniques"][rest_columns][
            ~unique_columns
        ].index
        return column_info

    if df_counts is None:
        df_counts = df.count()
        df_counts.name = "counts"
    uniques = get_df_uniques(df)
    missing = get_df_missing(df=df, df_counts=df_counts)
    stats = pd.concat([df_counts, uniques, missing], axis=1, sort=True)

    # settings types
    stats["types"] = ""
    columns_info = get_df_columns_info()
    for ctype, columns in columns_info.items():
        stats.loc[columns, "types"] = ctype
    return stats.transpose()[df.columns]


def get_df_columns_types(columns_stats: pd.DataFrame):
    return pd.value_counts(columns_stats.loc["types"])


def get_deviation_of_mean(
    series: pd.Series, multiplier: int = 3, df_length: int = None
) -> Tuple[int, float]:
    """
    Returns count of values deviating of the mean, i.e. larger than `multiplier` * `std`.
    Args:
        series: Series to perform operation over.
        multiplier: The value to use as
    Returns:
        tuple
    """
    if df_length is None:
        df_length = len(series)
    capped_series = np.minimum(series, series.mean() + multiplier * series.std())
    count = pd.value_counts(series != capped_series)
    count = count[True] if True in count else 0
    return count, to_percentage(count / df_length)


def get_median_absolute_deviation(
    series, multiplier=3, df_length: int = None
) -> Tuple[int, float]:
    """
    Returns count of values larger than `multiplier` * `mad`
    Args:
        series: Series to perform operation over.
        multiplier: The value to use as
    Returns:
        tuple
    """
    if df_length is None:
        df_length = len(series)
    capped_series = np.minimum(series, series.median() + multiplier * series.mad())
    count = pd.value_counts(series != capped_series)
    count = count[True] if True in count else 0
    return count, to_percentage(count / df_length)


def get_top_correlations(
    df: pd.DataFrame, column: str, threshold: float = 0.65, top: int = 3, df_corr=None
) -> Dict:
    """
    Returns count of values larger than `multiplier` * `mad`
    Args:
        df: dataframe.
        column: name of the column to calculate correlation for.
    Returns:
        tuple
    """
    if df_corr is None:
        df_corr = df.corr()
    column_corr = np.fabs(df_corr[column].drop(column)).sort_values(
        ascending=False, inplace=False
    )
    top_corr = column_corr[(column_corr > threshold)][:top].index
    return df_corr[column][top_corr].to_dict()


def get_top_correlations_description(
    df: pd.DataFrame, column: str, threshold: float = 0.65, top: int = 3, df_corr=None
) -> str:
    correlations = get_top_correlations(
        df=df, column=column, threshold=threshold, top=top, df_corr=df_corr
    )
    return ", ".join(
        "{}: {}".format(col, to_percentage(val)) for col, val in correlations.items()
    )


def get_numeric_summary(
    df: pd.DataFrame,
    column: str,
    columns_stats: pd.DataFrame = None,
    df_length: int = None,
    plot: bool = False,
):
    series = df[column]

    if plot:
        try:
            series.hist()
        except ImportError:
            pass

    if df_length is None:
        df_length = len(series)

    if columns_stats is None:
        columns_stats = get_df_column_stats(df[[column]])

    stats = {
        "mean": series.mean(),
        "std": series.std(),
        "variance": series.var(),
        "min": series.min(),
        "max": series.max(),
        "mode": series.mode()[0],
    }

    for x in np.array([0.05, 0.25, 0.5, 0.75, 0.95]):
        stats[to_percentage(x)] = series.quantile(x)

    stats["iqr"] = stats["75%"] - stats["25%"]
    stats["kurtosis"] = series.kurt()
    stats["skewness"] = series.skew()
    stats["sum"] = series.sum()
    stats["mad"] = series.mad()
    stats["cv"] = stats["std"] / stats["mean"] if stats["mean"] else np.nan
    stats["zeros_num"] = df_length - np.count_nonzero(series)
    stats["zeros_perc"] = to_percentage(stats["zeros_num"] / df_length)
    deviation_of_mean, deviation_of_mean_perc = get_deviation_of_mean(
        series, df_length=df_length
    )
    stats["deviating_of_mean"] = deviation_of_mean
    stats["deviating_of_mean_perc"] = deviation_of_mean_perc
    (
        deviating_of_median,
        deviating_of_median_perc,
    ) = get_median_absolute_deviation(series, df_length=df_length)
    stats["deviating_of_median"] = deviating_of_median
    stats["deviating_of_median_perc"] = deviating_of_median_perc
    stats["top_correlations"] = get_top_correlations_description(df=df, column=column)
    return pd.concat([pd.Series(stats, name=column), columns_stats[column]], sort=True)


def get_date_summary(df: pd.DataFrame, column: str, columns_stats: pd.DataFrame = None):
    if columns_stats is None:
        columns_stats = get_df_column_stats(df[[column]])

    series = df[column]
    stats = {"min": series.min(), "max": series.max()}
    stats["range"] = stats["max"] - stats["min"]
    return pd.concat([pd.Series(stats, name=column), columns_stats[column]], sort=True)


def get_categorical_summary(
    df: pd.DataFrame, column: str, columns_stats: pd.DataFrame = None
):
    if columns_stats is None:
        columns_stats = get_df_column_stats(df[[column]])

    series = df[column]
    # Only run if at least 1 non-missing value
    value_counts = series.value_counts()
    stats = {
        "top": "{}: {}".format(value_counts.index[0], value_counts.iloc[0]),
    }
    return pd.concat([pd.Series(stats, name=column), columns_stats[column]], sort=True)


def get_constant_summary(df: pd.DataFrame, column: str):
    return "This is a constant value: {}".format(df[column][0])


def get_bool_summary(
    df: pd.DataFrame,
    column: str,
    columns_stats: pd.DataFrame = None,
    df_length: int = None,
):
    series = df[column]

    if columns_stats is None:
        columns_stats = get_df_column_stats(df[[column]])

    if df_length is None:
        df_length = len(series)

    stats = {}
    for class_name, class_value in sorted(dict(series.value_counts()).items()):
        stats['"{}" count'.format(class_name)] = "{}".format(class_value)
        stats['"{}" perc'.format(class_name)] = "{}".format(
            to_percentage(class_value / df_length)
        )

    return pd.concat([pd.Series(stats, name=column), columns_stats[column]], sort=True)


def get_unique_summary(
    df: pd.DataFrame, column: str, columns_stats: pd.DataFrame = None
):
    if columns_stats is None:
        columns_stats = get_df_column_stats(df[[column]])

    return columns_stats[column]


def get_df_column_summary(
    df: pd.DataFrame,
    column: str,
    columns_stats: pd.DataFrame = None,
    df_length: int = None,
    plot: bool = False,
):
    if columns_stats is None:
        columns_stats = get_df_column_stats(df[[column]])

    if df_length is None:
        df_length = len(df)

    column_type = columns_stats.loc["types"][column]
    if column_type == DF_TYPE_NUMERIC:
        return get_numeric_summary(
            df=df,
            column=column,
            columns_stats=columns_stats,
            df_length=df_length,
            plot=plot,
        )
    if column_type == DF_TYPE_CATEGORICAL:
        return get_categorical_summary(
            df=df, column=column, columns_stats=columns_stats
        )
    if column_type == DF_TYPE_BOOL:
        return get_bool_summary(
            df=df, column=column, columns_stats=columns_stats, df_length=df_length
        )
    if column_type == DF_TYPE_UNIQUE:
        return get_unique_summary(df=df, column=column, columns_stats=columns_stats)
    if column_type == DF_TYPE_DATE:
        return get_date_summary(df=df, column=column, columns_stats=columns_stats)
    if column_type == DF_TYPE_CONSTANT:
        return get_constant_summary(df=df, column=column)



from .renderer import PlotlyRenderer
from .mplexporter import Exporter


def mpl_to_plotly(fig, resize=False, strip_style=False, verbose=False):
    """Convert a matplotlib figure to plotly dictionary and send.
    All available information about matplotlib visualizations are stored
    within a matplotlib.figure.Figure object. You can create a plot in python
    using matplotlib, store the figure object, and then pass this object to
    the fig_to_plotly function. In the background, mplexporter is used to
    crawl through the mpl figure object for appropriate information. This
    information is then systematically sent to the PlotlyRenderer which
    creates the JSON structure used to make plotly visualizations. Finally,
    these dictionaries are sent to plotly and your browser should open up a
    new tab for viewing! Optionally, if you're working in IPython, you can
    set notebook=True and the PlotlyRenderer will call plotly.iplot instead
    of plotly.plot to have the graph appear directly in the IPython notebook.
    Note, this function gives the user access to a simple, one-line way to
    render an mpl figure in plotly. If you need to trouble shoot, you can do
    this step manually by NOT running this fuction and entereing the following:
    ===========================================================================
    from plotly.matplotlylib import mplexporter, PlotlyRenderer
    # create an mpl figure and store it under a varialble 'fig'
    renderer = PlotlyRenderer()
    exporter = mplexporter.Exporter(renderer)
    exporter.run(fig)
    ===========================================================================
    You can then inspect the JSON structures by accessing these:
    renderer.layout -- a plotly layout dictionary
    renderer.data -- a list of plotly data dictionaries
    """

    # Update vendor
    # This code was taken from:
    # https://github.com/matplotlib/matplotlib/pull/16772/files#diff-506cc6d736a0593e8bb820981b2c12ae # noqa
    # Removed in https://github.com/matplotlib/matplotlib/pull/16772
    from matplotlib.spines import Spine

    def is_frame_like(self):
        """return True if directly on axes frame
        This is useful for determining if a spine is the edge of an
        old style MPL plot. If so, this function will return True.
        """
        self._ensure_position_is_set()
        position = self._position
        if isinstance(position, str):
            if position == "center":
                position = ("axes", 0.5)
            elif position == "zero":
                position = ("data", 0)
        if len(position) != 2:
            raise ValueError("position should be 2-tuple")
        position_type, amount = position
        if position_type == "outward" and amount == 0:
            return True
        else:
            return False

    Spine.is_frame_like = is_frame_like

    renderer = PlotlyRenderer()
    Exporter(renderer).run(fig)
    if resize:
        renderer.resize()
    if strip_style:
        renderer.strip_style()
    if verbose:
        print(renderer.msg)
    return renderer.plotly_fig



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from traceml import tracking


def callback(run=None):
    run = tracking.get_or_create_run(run)

    def _callback(env):
        res = {}
        for data_name, eval_name, value, _ in env.evaluation_result_list:
            key = data_name + "-" + eval_name
            res[key] = value
        run.log_metrics(step=env.iteration, **res)

    return _callback



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import ujson

from traceml import tracking
from traceml.exceptions import TracemlException
from traceml.logger import logger

try:
    import xgboost as xgb

    from xgboost import Booster
except ImportError:
    raise TracemlException("xgboost is required to use the tracking callback")


if TYPE_CHECKING:
    from traceml.tracking import Run


def _get_cv(model):
    return getattr(model, "cvfolds", False)


def _log_importance(run, model, model_folds, max_num_features, **kwargs):
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Please install matplotlib to log importance")

    if model_folds:
        for i, fold in enumerate(model_folds):
            importance = xgb.plot_importance(
                fold.bst, max_num_features=max_num_features, **kwargs
            )
            run.log_mpl_plotly_chart(
                name="feature_importance", figure=importance.figure, step=i
            )
    else:
        importance = xgb.plot_importance(
            model, max_num_features=max_num_features, **kwargs
        )
        run.log_mpl_plotly_chart(name="feature_importance", figure=importance.figure)
    plt.close("all")


def _log_model(run, model, model_folds):
    def _save(file_model, file_name):
        asset_path = run.get_outputs_path(file_name)
        file_model.save_model(asset_path)
        run.log_model_ref(asset_path, framework="xgboost")

    if model_folds:
        for i, cvpack in enumerate(model_folds):
            _save(cvpack.bst, "model-{}".format(i))
    else:  # train case
        _save(model, "model")


def callback(
    log_model: bool = True,
    log_importance: bool = True,
    max_num_features: int = None,
    run: "Run" = None,
):
    run = tracking.get_or_create_run(run)

    def callback(env):
        # Log metrics after iteration
        metrics = {}
        for item in env.evaluation_result_list:
            if len(item) == 2:  # train case
                metrics[item[0]] = item[1]
            if len(item) == 3:  # cv case
                metrics["{}-mean".format(item[0])] = item[1]
                metrics["{}-std".format(item[0])] = item[2]
        run.log_metrics(
            **metrics,
            step=env.iteration,
        )

        model = getattr(env, "model")
        model_folds = _get_cv(env)

        # Log booster, end of training
        if log_model:
            _log_model(run=run, model=model, model_folds=model_folds)

        # Log feature importance, end of training
        if env.iteration + 1 == env.end_iteration and log_importance:
            try:
                _log_importance(
                    run,
                    model=model,
                    model_folds=model_folds,
                    max_num_features=max_num_features,
                )
            except Exception as e:
                logger.info("Failed logging feature importance %s", e)

    return callback


class Callback(xgb.callback.TrainingCallback):
    def __init__(
        self,
        run: "Run" = None,
        log_model: bool = True,
        log_importance: bool = True,
        importance_type: str = "gain",
        max_num_features: int = None,
    ):

        self.log_model: bool = log_model
        self.log_importance: bool = log_importance
        self.importance_type: str = importance_type
        self.max_num_features: int = max_num_features
        self.run = tracking.get_or_create_run(run)

    def after_training(self, model: Booster) -> Booster:
        model_folds = _get_cv(model)

        if self.log_model:
            _log_model(run=self.run, model=model, model_folds=model_folds)

        if self.log_importance:
            _log_importance(
                self.run,
                model=model,
                model_folds=model_folds,
                max_num_features=self.max_num_features,
            )

        if model_folds:
            config = {}
            for i, fold in enumerate(model_folds):
                config["fold_{}_config".format(i)] = ujson.loads(fold.bst.save_config())
            if config:
                self.run.log_inputs(**config)
        else:
            self.run.log_inputs(config=ujson.loads(model.save_config()))
            outputs = {}
            if "best_score" in model.attributes().keys():
                outputs["best_score"] = model.attributes()["best_score"]
            if "best_iteration" in model.attributes().keys():
                outputs["best_iteration"] = model.attributes()["best_iteration"]
            self.run.log_outputs(**outputs)

        return model

    def after_iteration(self, model: Booster, epoch: int, evals_log: dict) -> bool:
        metrics = {}
        for stage, metrics_dict in evals_log.items():
            for metric_name, metric_values in evals_log[stage].items():
                if _get_cv(model):
                    mean, std = metric_values[-1]
                    metrics["{}-{}-mean".format(stage, metric_name)] = mean
                    metrics["{}-{}-std".format(stage, metric_name)] = std
                else:
                    metrics["{}-{}".format(stage, metric_name)] = metric_values[-1]

        if metrics:
            self.run.log_metrics(step=epoch, **metrics)

        return False



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from polyaxon.utils.path_utils import copy_file_or_dir_path
from traceml.events import V1EventArtifact

try:
    import numpy as np
except ImportError:
    np = None


def artifact_path(
    from_path: str, asset_path: str, kind: str, asset_rel_path: str = None
) -> V1EventArtifact:
    copy_file_or_dir_path(from_path, asset_path)
    return V1EventArtifact(kind=kind, path=asset_rel_path or asset_path)



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union


def _sanitize_value(value: Union[str, int, float]) -> Union[int, float]:
    fvalue = float(value)
    ivalue = int(fvalue)
    return ivalue if ivalue == fvalue else fvalue


def to_cpu_value(cpu_definition: Union[str, int, float]) -> float:
    try:
        return float(cpu_definition)
    except (ValueError, TypeError):
        pass

    cpu_definition = cpu_definition.lower()
    cpu_unit = cpu_definition[-1]
    cpu_value = cpu_definition[:-1]
    if cpu_unit == "m":
        cpu = _sanitize_value(cpu_value) / 1000
    elif cpu_unit == "u":
        cpu = _sanitize_value(cpu_value) / 1000**2
    elif cpu_unit == "n":
        cpu = _sanitize_value(cpu_value) / 1000**3
    else:
        cpu = cpu_definition
    return _sanitize_value(cpu)


def to_memory_bytes(mem_definition: Union[str, int, float]) -> int:
    try:
        return int(float(mem_definition))
    except (ValueError, TypeError):
        pass

    def _get_value(unit, value, multiplier):
        if unit in multiplier.keys():
            return _sanitize_value(value) * multiplier.get(unit, 1)

    fixed_point_unit_multiplier = {
        "k": 1000,
        "m": 1000**2,
        "g": 1000**3,
        "t": 1000**4,
        "p": 1000**5,
        "e": 1000**6,
    }

    power_two_unit_multiplier = {
        "ki": 1024,
        "mi": 1024**2,
        "gi": 1024**3,
        "ti": 1024**4,
        "pi": 1024**5,
        "ei": 1024**6,
    }

    mem_definition = mem_definition.lower()

    mem_unit = mem_definition[-2:]
    mem_value = mem_definition[:-2]
    memory = _get_value(mem_unit, mem_value, power_two_unit_multiplier)
    if memory is not None:
        return memory

    mem_unit = mem_definition[-1:]
    mem_value = mem_definition[:-1]
    memory = _get_value(mem_unit, mem_value, fixed_point_unit_multiplier)
    if memory is not None:
        return memory

    return 0


def to_unit_memory(number, precision: int = 2):
    """Creates a string representation of memory size given `number`."""
    kb = 1024

    number /= kb

    if number < 100:
        return "{} Ki".format(round(number, precision))

    number /= kb
    if number < 300:
        return "{} Mi".format(round(number, precision))

    number /= kb
    if number < 900:
        return "{} Gi".format(round(number, precision))

    number /= kb
    if number < 900:
        return "{} Ti".format(round(number, precision))

    number /= kb
    if number < 900:
        return "{} Pi".format(round(number, precision))

    number /= kb
    return "{} Ei".format(round(number, precision))


def number_percentage_format(x, precision: int = None, use_comma: bool = False):
    if precision is None:
        return x
    eps = 0.000000001
    comma = "," if use_comma else ""
    num_format = (
        "{{0:{}.0f}}".format(comma)
        if abs(int(x) - x) < eps
        else "{{0:{}.{}f}}".format(comma, precision)
    )
    return num_format.format(x)


def to_percentage(
    number, rounding: int = 2, precision: int = None, use_comma: bool = False
):
    """Creates a percentage string representation from the given `number`. The
    number is multiplied by 100 before adding a '%' character.

    Raises `ValueError` if `number` cannot be converted to a number.
    """
    number = float(number) * 100
    number_as_int = int(number)
    rounded = round(number, rounding)
    value = (
        number_as_int
        if number_as_int == rounded
        else number_percentage_format(rounded, precision, use_comma)
    )

    return "{}%".format(value)


def format_sizeof(num, suffix="B"):
    """
    Print in human friendly format
    """
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Yi", suffix)



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.constants.globals import UNKNOWN
from polyaxon.utils.np_utils import to_np
from polyaxon.utils.path_utils import check_or_create_path, copy_file_path
from traceml.events import V1EventAudio
from traceml.logger import logger
from traceml.processors.errors import NUMPY_ERROR_MESSAGE

try:
    import numpy as np
except ImportError:
    np = None


def audio_path(
    from_path: str, asset_path: str, content_type=None, asset_rel_path: str = None
) -> V1EventAudio:
    copy_file_path(from_path, asset_path)
    return V1EventAudio(path=asset_rel_path or asset_path, content_type=content_type)


def audio(asset_path: str, tensor, sample_rate=44100, asset_rel_path: str = None):
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    tensor = to_np(tensor)
    tensor = tensor.squeeze()
    if abs(tensor).max() > 1:
        print("warning: audio amplitude out of range, auto clipped.")
        tensor = tensor.clip(-1, 1)
    assert tensor.ndim == 1, "input tensor should be 1 dimensional."

    tensor_list = [int(32767.0 * x) for x in tensor]

    import struct
    import wave

    check_or_create_path(asset_path, is_dir=False)

    wave_write = wave.open(asset_path, "wb")
    wave_write.setnchannels(1)
    wave_write.setsampwidth(2)
    wave_write.setframerate(sample_rate)
    tensor_enc = b""
    for v in tensor_list:
        tensor_enc += struct.pack("<h", v)

    wave_write.writeframes(tensor_enc)
    wave_write.close()
    return V1EventAudio(
        sample_rate=sample_rate,
        num_channels=1,
        length_frames=len(tensor_list),
        path=asset_rel_path or asset_path,
        content_type="audio/wav",
    )



"""
matplotlylib
============

This module converts matplotlib figure objects into JSON structures which can
be understood and visualized by Plotly.

Most of the functionality should be accessed through the parent directory's
'tools' module or 'plotly' package.

"""
from __future__ import absolute_import

from .renderer import PlotlyRenderer
from .mplexporter import Exporter
from .tools import mpl_to_plotly



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.constants.globals import UNKNOWN
from polyaxon.utils.np_utils import calculate_scale_factor, to_np
from polyaxon.utils.path_utils import check_or_create_path, copy_file_path
from traceml.events import V1EventVideo
from traceml.logger import logger
from traceml.processors.errors import MOVIEPY_ERROR_MESSAGE, NUMPY_ERROR_MESSAGE

try:
    import numpy as np
except ImportError:
    np = None


def video_path(
    from_path: str, asset_path: str, content_type=None, asset_rel_path: str = None
) -> V1EventVideo:
    copy_file_path(from_path, asset_path)
    return V1EventVideo(path=asset_rel_path or asset_path, content_type=content_type)


def video(
    asset_path: str, tensor, fps=4, content_type="gif", asset_rel_path: str = None
):
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    tensor = to_np(tensor)
    tensor = prepare_video(tensor)
    # If user passes in uint8, then we don't need to rescale by 255
    scale_factor = calculate_scale_factor(tensor)
    tensor = tensor.astype(np.float32)
    tensor = (tensor * scale_factor).astype(np.uint8)
    return make_video(
        asset_path, tensor, fps, content_type, asset_rel_path=asset_rel_path
    )


def make_video(
    asset_path: str, tensor, fps, content_type="gif", asset_rel_path: str = None
):
    try:
        import moviepy  # noqa: F401
    except ImportError:
        logger.warning(MOVIEPY_ERROR_MESSAGE)
        return UNKNOWN
    try:
        from moviepy import editor as mpy
    except ImportError:
        logger.warning(
            "moviepy is installed, but can't import moviepy.editor.",
            "Some packages could be missing [imageio, requests]",
        )
        return

    t, h, w, c = tensor.shape

    # encode sequence of images into gif string
    clip = mpy.ImageSequenceClip(list(tensor), fps=fps)

    check_or_create_path(asset_path, is_dir=False)

    try:  # older version of moviepy
        if content_type == "gif":
            clip.write_gif(asset_path, verbose=False, progress_bar=False)
        else:
            clip.write_videofile(asset_path, verbose=False, progress_bar=False)
    except TypeError:
        if content_type == "gif":
            clip.write_gif(asset_path, verbose=False)
        else:
            clip.write_videofile(asset_path, verbose=False)

    return V1EventVideo(
        height=h,
        width=w,
        colorspace=c,
        path=asset_rel_path or asset_path,
        content_type=content_type,
    )


def prepare_video(data):
    """
    Converts a 5D tensor [batchsize, time(frame), channel(color), height, width]
    into 4D tensor with dimension [time(frame), new_width, new_height, channel].
    A batch of images are spreaded to a grid, which forms a frame.
    e.g. Video with batchsize 16 will have a 4x4 grid.
    """
    if not np:
        logger.warning(NUMPY_ERROR_MESSAGE)
        return UNKNOWN

    b, t, c, h, w = data.shape

    if data.dtype == np.uint8:
        data = np.float32(data) / 255.0

    def is_power2(num):
        return num != 0 and ((num & (num - 1)) == 0)

    # pad to nearest power of 2, all at once
    if not is_power2(data.shape[0]):
        len_addition = int(2 ** data.shape[0].bit_length() - data.shape[0])
        data = np.concatenate(
            (data, np.zeros(shape=(len_addition, t, c, h, w))), axis=0
        )

    n_rows = 2 ** ((b.bit_length() - 1) // 2)
    n_cols = data.shape[0] // n_rows

    data = np.reshape(data, newshape=(n_rows, n_cols, t, c, h, w))
    data = np.transpose(data, axes=(2, 0, 4, 1, 5, 3))
    return np.reshape(data, newshape=(t, n_rows * h, n_cols * w, c))



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NUMPY_ERROR_MESSAGE = "numpy is required for this tracking operation!"
PANDAS_ERROR_MESSAGE = "pandas is required for this tracking operation!"
PIL_ERROR_MESSAGE = "PIL/Pillow is required for this tracking operation!"
MOVIEPY_ERROR_MESSAGE = "moviepy is required for this tracking operation!"
MATPLOTLIB_ERROR_MESSAGE = "matplotlib is required for this tracking operation!"
PLOTLY_ERROR_MESSAGE = "plotly is required for this tracking operation!"
BOKEH_ERROR_MESSAGE = "bokeh is required for this tracking operation!"
SKLEARN_ERROR_MESSAGE = "sklearn is required for this tracking operation!"



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence, Union

from polyaxon.client import RunClient
from traceml.artifacts import V1ArtifactKind, V1RunArtifact
from traceml.tracking.run import Run

TRACKING_RUN: Run = None


def init(
    owner: str = None,
    project: str = None,
    run_uuid: str = None,
    client: RunClient = None,
    track_code: bool = True,
    track_env: bool = True,
    refresh_data: bool = False,
    artifacts_path: str = None,
    collect_artifacts: str = None,
    collect_resources: str = None,
    is_offline: bool = None,
    is_new: bool = None,
    name: str = None,
    description: str = None,
    tags: List[str] = None,
) -> Optional[Run]:
    """Tracking module is similar to the tracking client without the need to create a run instance.

    The tracking module allows you to call all tracking methods directly from the top level module.

    This could be very convenient especially if you are running in-cluster experiments:

    ```python
    from polyaxon import tracking

    tracking.init()
    ...
    tracking.log_metrics(step=1, loss=0.09, accuracy=0.75)
    ...
    tracking.log_metrics(step=1, loss=0.02, accuracy=0.85)
    ...
    ```

    > A global `TRACKING_RUN` will be set on the module.


        Args:
            owner: str, optional, the owner is the username or
                 the organization name owning this project.
            project: str, optional, project name owning the run(s).
            run_uuid: str, optional, run uuid.
            client: [PolyaxonClient](/docs/core/python-library/polyaxon-client/), optional,
                 an instance of a configured client, if not passed,
                 a new instance will be created based on the available environment.
            track_code: bool, optional, default True, to track code version.
                 Polyaxon will try to track information about any repo
                 configured in the context where this client is instantiated.
            track_env: bool, optional, default True, to track information about the environment.
            refresh_data: bool, optional, default False, to refresh the run data at instantiation.
            refresh_data: bool, optional, default False, to instruct the run to resume,
                 only useful when the run is not managed by Polyaxon.
            artifacts_path: str, optional, for in-cluster runs it will be set automatically.
            collect_artifacts: bool, optional,
                 similar to the env var flag `POLYAXON_COLLECT_ARTIFACTS`, this env var is `True`
                 by default for managed runs and is controlled by the plugins section.
            collect_resources: bool, optional,
                 similar to the env var flag `POLYAXON_COLLECT_RESOURCES`, this env var is `True`
                 by default for managed runs and is controlled by the plugins section.
            is_offline: bool, optional,
                 To trigger the offline mode manually instead of depending on `POLYAXON_IS_OFFLINE`.
            is_new: bool, optional,
                 Force the creation of a new run instead of trying to discover a cached run or
                 refreshing an instance from the env var
            name: str, optional,
                 When `is_new` or `is_offline` is set to true, a new instance is created and
                 you can initialize that new run with a name.
            description: str, optional,
                 When `is_new` or `is_offline` is set to true, a new instance is created and
                 you can initialize that new run with a description.
            tags: str or List[str], optional,
                 When `is_new` or `is_offline` is set to true, a new instance is created and
                 you can initialize that new run with tags.

        Raises:
            PolyaxonClientException: If no owner and/or project are passed and Polyaxon cannot
                 resolve the values from the environment.
    """
    global TRACKING_RUN

    TRACKING_RUN = Run(
        owner=owner,
        project=project,
        run_uuid=run_uuid,
        client=client,
        track_code=track_code,
        refresh_data=refresh_data,
        track_env=track_env,
        artifacts_path=artifacts_path,
        collect_artifacts=collect_artifacts,
        collect_resources=collect_resources,
        is_offline=is_offline,
        is_new=is_new,
        name=name,
        description=description,
        tags=tags,
    )
    return TRACKING_RUN


def get_or_create_run(tracking_run: Run = None) -> Optional[Run]:
    """Get or create a new tracking run.

    It tries to create a new instance, for in-cluster runs, this will work automatically.

    This is used inside some Polyaxon callbacks, you should use `init` instead.
    """
    global TRACKING_RUN

    if tracking_run:
        return tracking_run
    if TRACKING_RUN:
        return TRACKING_RUN

    init()
    return TRACKING_RUN


def get_artifacts_path(
    rel_path: str = None,
    ensure_path: bool = False,
    is_dir: bool = False,
    use_store_path: bool = False,
):
    global TRACKING_RUN
    return TRACKING_RUN.get_artifacts_path(
        rel_path=rel_path,
        ensure_path=ensure_path,
        is_dir=is_dir,
        use_store_path=use_store_path,
    )


get_artifacts_path.__doc__ = Run.get_artifacts_path.__doc__


def get_outputs_path(
    rel_path: str = None,
    ensure_path: bool = True,
    is_dir: bool = False,
    use_store_path: bool = False,
):
    global TRACKING_RUN
    return TRACKING_RUN.get_outputs_path(
        rel_path=rel_path,
        ensure_path=ensure_path,
        is_dir=is_dir,
        use_store_path=use_store_path,
    )


get_outputs_path.__doc__ = Run.get_outputs_path.__doc__


def get_tensorboard_path(
    rel_path: str = "tensorboard",
    use_store_path: bool = False,
):
    global TRACKING_RUN
    return TRACKING_RUN.get_tensorboard_path(
        rel_path=rel_path, use_store_path=use_store_path
    )


get_tensorboard_path.__doc__ = Run.get_tensorboard_path.__doc__


def set_artifacts_path(artifacts_path: str, is_related: bool = False):
    global TRACKING_RUN
    TRACKING_RUN.set_artifacts_path(artifacts_path, is_related)


set_artifacts_path.__doc__ = Run.set_artifacts_path.__doc__


def set_run_event_logger():
    global TRACKING_RUN
    TRACKING_RUN.set_run_event_logger()


set_run_event_logger.__doc__ = Run.set_run_event_logger.__doc__


def set_run_resource_logger():
    global TRACKING_RUN
    TRACKING_RUN.set_run_resource_logger()


set_run_resource_logger.__doc__ = Run.set_run_resource_logger.__doc__


def set_run_process_sidecar():
    global TRACKING_RUN
    TRACKING_RUN.set_run_process_sidecar()


set_run_process_sidecar.__doc__ = Run.set_run_process_sidecar.__doc__


def log_metric(name: str, value: float, step: int = None, timestamp: datetime = None):
    global TRACKING_RUN
    TRACKING_RUN.log_metric(
        name=name,
        value=value,
        step=step,
        timestamp=timestamp,
    )


log_metric.__doc__ = Run.log_metric.__doc__


def log_metrics(step: int = None, timestamp: datetime = None, **metrics):
    global TRACKING_RUN
    TRACKING_RUN.log_metrics(step=step, timestamp=timestamp, **metrics)


log_metrics.__doc__ = Run.log_metrics.__doc__


def log_roc_auc_curve(
    name: str, fpr, tpr, auc: float = None, step: int = None, timestamp: datetime = None
):
    global TRACKING_RUN
    TRACKING_RUN.log_roc_auc_curve(
        name=name,
        fpr=fpr,
        tpr=tpr,
        auc=auc,
        step=step,
        timestamp=timestamp,
    )


log_roc_auc_curve.__doc__ = Run.log_roc_auc_curve.__doc__


def log_sklearn_roc_auc_curve(
    name: str,
    y_preds,
    y_targets,
    step: int = None,
    timestamp: datetime = None,
    is_multi_class: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.log_sklearn_roc_auc_curve(
        name=name,
        y_preds=y_preds,
        y_targets=y_targets,
        step=step,
        timestamp=timestamp,
        is_multi_class=is_multi_class,
    )


log_sklearn_roc_auc_curve.__doc__ = Run.log_sklearn_roc_auc_curve.__doc__


def log_pr_curve(
    name: str,
    precision,
    recall,
    average_precision=None,
    step: int = None,
    timestamp: datetime = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_pr_curve(
        name=name,
        precision=precision,
        recall=recall,
        average_precision=average_precision,
        step=step,
        timestamp=timestamp,
    )


log_pr_curve.__doc__ = Run.log_pr_curve.__doc__


def log_sklearn_pr_curve(
    name: str,
    y_preds,
    y_targets,
    step: int = None,
    timestamp: datetime = None,
    is_multi_class: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.log_sklearn_pr_curve(
        name=name,
        y_preds=y_preds,
        y_targets=y_targets,
        step=step,
        timestamp=timestamp,
        is_multi_class=is_multi_class,
    )


log_sklearn_pr_curve.__doc__ = Run.log_sklearn_pr_curve.__doc__


def log_curve(
    name: str,
    x,
    y,
    annotation: str = None,
    step: int = None,
    timestamp: datetime = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_curve(
        name=name,
        x=x,
        y=y,
        annotation=annotation,
        step=step,
        timestamp=timestamp,
    )


log_curve.__doc__ = Run.log_curve.__doc__


def log_confusion_matrix(
    name: str,
    x,
    y,
    z=None,
    step: int = None,
    timestamp: datetime = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_confusion_matrix(
        name=name,
        x=x,
        y=y,
        z=z,
        step=step,
        timestamp=timestamp,
    )


log_confusion_matrix.__doc__ = Run.log_confusion_matrix.__doc__


def log_image(
    data: Any,
    name: str = None,
    step: int = None,
    timestamp: datetime = None,
    rescale=1,
    dataformats: str = "CHW",
    ext: str = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_image(
        data=data,
        name=name,
        step=step,
        timestamp=timestamp,
        rescale=rescale,
        dataformats=dataformats,
        ext=ext,
    )


log_image.__doc__ = Run.log_image.__doc__


def log_image_with_boxes(
    tensor_image,
    tensor_boxes,
    name: str = None,
    step: int = None,
    timestamp: datetime = None,
    rescale: int = 1,
    dataformats: str = "CHW",
):
    global TRACKING_RUN
    TRACKING_RUN.log_image_with_boxes(
        tensor_image=tensor_image,
        tensor_boxes=tensor_boxes,
        name=name,
        step=step,
        timestamp=timestamp,
        rescale=rescale,
        dataformats=dataformats,
    )


log_image_with_boxes.__doc__ = Run.log_image_with_boxes.__doc__


def log_mpl_image(
    data,
    name: str = None,
    close: bool = True,
    step: int = None,
    timestamp: datetime = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_mpl_image(
        data=data,
        name=name,
        close=close,
        step=step,
        timestamp=timestamp,
    )


log_mpl_image.__doc__ = Run.log_mpl_image.__doc__


def log_video(
    data,
    name: str = None,
    fps: int = 4,
    step: int = None,
    timestamp: datetime = None,
    content_type: str = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_video(
        data=data,
        name=name,
        fps=fps,
        step=step,
        timestamp=timestamp,
        content_type=content_type,
    )


log_video.__doc__ = Run.log_video.__doc__


def log_audio(
    data,
    name: str = None,
    sample_rate: int = 44100,
    step: int = None,
    timestamp: datetime = None,
    content_type: str = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_audio(
        data=data,
        name=name,
        sample_rate=sample_rate,
        step=step,
        timestamp=timestamp,
        content_type=content_type,
    )


log_audio.__doc__ = Run.log_audio.__doc__


def log_text(name: str, text: str, step: int = None, timestamp: datetime = None):
    global TRACKING_RUN
    TRACKING_RUN.log_text(
        name=name,
        text=text,
        step=step,
        timestamp=timestamp,
    )


log_text.__doc__ = Run.log_text.__doc__


def log_html(name: str, html: str, step: int = None, timestamp: datetime = None):
    global TRACKING_RUN
    TRACKING_RUN.log_html(
        name=name,
        html=html,
        step=step,
        timestamp=timestamp,
    )


log_html.__doc__ = Run.log_html.__doc__


def log_np_histogram(
    name: str, values, counts, step: int = None, timestamp: datetime = None
):
    global TRACKING_RUN
    TRACKING_RUN.log_np_histogram(
        name=name,
        values=values,
        counts=counts,
        step=step,
        timestamp=timestamp,
    )


log_np_histogram.__doc__ = Run.log_np_histogram.__doc__


def log_histogram(
    name: str, values, bins, max_bins=None, step: int = None, timestamp: datetime = None
):
    global TRACKING_RUN
    TRACKING_RUN.log_histogram(
        name=name,
        values=values,
        bins=bins,
        max_bins=max_bins,
        step=step,
        timestamp=timestamp,
    )


log_histogram.__doc__ = Run.log_histogram.__doc__


def log_model(
    path: str,
    name: str = None,
    framework: str = None,
    summary: Dict = None,
    step: int = None,
    timestamp: datetime = None,
    rel_path: str = None,
    skip_hash_calculation: bool = False,
    **kwargs,
):
    global TRACKING_RUN
    TRACKING_RUN.log_model(
        path=path,
        name=name,
        framework=framework,
        summary=summary,
        step=step,
        timestamp=timestamp,
        rel_path=rel_path,
        skip_hash_calculation=skip_hash_calculation,
        **kwargs,
    )


log_model.__doc__ = Run.log_model.__doc__


def log_artifact(
    path: str,
    name: str = None,
    kind: str = None,
    summary: Dict = None,
    step: int = None,
    timestamp: datetime = None,
    rel_path: str = None,
    skip_hash_calculation: bool = False,
    **kwargs,
):
    global TRACKING_RUN
    TRACKING_RUN.log_artifact(
        path=path,
        name=name,
        kind=kind,
        summary=summary,
        step=step,
        timestamp=timestamp,
        rel_path=rel_path,
        skip_hash_calculation=skip_hash_calculation,
        **kwargs,
    )


log_artifact.__doc__ = Run.log_artifact.__doc__


def log_dataframe(
    df,
    name: str,
    content_type: str = V1ArtifactKind.CSV,
    step: int = None,
    timestamp: datetime = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_dataframe(
        df=df,
        name=name,
        content_type=content_type,
        step=step,
        timestamp=timestamp,
    )


log_dataframe.__doc__ = Run.log_dataframe.__doc__


def log_plotly_chart(name: str, figure, step: int = None, timestamp: datetime = None):
    global TRACKING_RUN
    TRACKING_RUN.log_plotly_chart(
        name=name,
        figure=figure,
        step=step,
        timestamp=timestamp,
    )


log_plotly_chart.__doc__ = Run.log_plotly_chart.__doc__


def log_bokeh_chart(name: str, figure, step: int = None, timestamp: datetime = None):
    global TRACKING_RUN
    TRACKING_RUN.log_bokeh_chart(
        name=name,
        figure=figure,
        step=step,
        timestamp=timestamp,
    )


log_bokeh_chart.__doc__ = Run.log_bokeh_chart.__doc__


def log_altair_chart(name: str, figure, step: int = None, timestamp: datetime = None):
    global TRACKING_RUN
    TRACKING_RUN.log_altair_chart(
        name=name,
        figure=figure,
        step=step,
        timestamp=timestamp,
    )


log_altair_chart.__doc__ = Run.log_altair_chart.__doc__


def log_mpl_plotly_chart(
    name: str,
    figure,
    step: int = None,
    timestamp: datetime = None,
    close: bool = True,
    fallback_to_image: bool = True,
):
    global TRACKING_RUN
    TRACKING_RUN.log_mpl_plotly_chart(
        name=name,
        figure=figure,
        step=step,
        timestamp=timestamp,
        close=close,
        fallback_to_image=fallback_to_image,
    )


log_mpl_plotly_chart.__doc__ = Run.log_mpl_plotly_chart.__doc__


def get_log_level():
    global TRACKING_RUN
    return TRACKING_RUN.get_log_level()


get_log_level.__doc__ = Run.get_log_level.__doc__


def set_description(description: str, async_req: bool = True):
    global TRACKING_RUN
    TRACKING_RUN.set_description(description=description, async_req=async_req)


set_description.__doc__ = Run.set_description.__doc__


def set_name(name: str, async_req: bool = True):
    global TRACKING_RUN
    TRACKING_RUN.set_name(name=name, async_req=async_req)


set_name.__doc__ = Run.set_name.__doc__


def end():
    global TRACKING_RUN
    TRACKING_RUN.end()


end.__doc__ = Run.end.__doc__


def log_status(
    status: str,
    reason: str = None,
    message: str = None,
    last_transition_time: datetime = None,
    last_update_time: datetime = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_status(
        status=status,
        reason=reason,
        message=message,
        last_transition_time=last_transition_time,
        last_update_time=last_update_time,
    )


log_status.__doc__ = Run.log_status.__doc__


def log_inputs(reset: bool = False, async_req: bool = True, **inputs):
    global TRACKING_RUN
    TRACKING_RUN.log_inputs(reset=reset, async_req=async_req, **inputs)


log_inputs.__doc__ = Run.log_inputs.__doc__


def log_outputs(reset: bool = False, async_req: bool = True, **outputs):
    global TRACKING_RUN
    TRACKING_RUN.log_outputs(reset=reset, async_req=async_req, **outputs)


log_outputs.__doc__ = Run.log_outputs.__doc__


def log_tags(
    tags: Union[str, Sequence[str]],
    reset: bool = False,
    async_req: bool = True,
):
    global TRACKING_RUN
    TRACKING_RUN.log_tags(tags=tags, reset=reset, async_req=async_req)


log_tags.__doc__ = Run.log_tags.__doc__


def log_meta(reset: bool = False, async_req: bool = True, **meta):
    global TRACKING_RUN
    TRACKING_RUN.log_meta(reset=reset, async_req=async_req, **meta)


log_meta.__doc__ = Run.log_meta.__doc__


def log_progress(value: float):
    global TRACKING_RUN
    TRACKING_RUN.log_progress(value=value)


log_progress.__doc__ = Run.log_progress.__doc__


def log_succeeded():
    global TRACKING_RUN
    TRACKING_RUN.log_succeeded()


log_succeeded.__doc__ = Run.log_succeeded.__doc__


def log_stopped():
    global TRACKING_RUN
    TRACKING_RUN.log_stopped()


log_stopped.__doc__ = Run.log_stopped.__doc__


def log_failed(reason: str = None, message: str = None):
    global TRACKING_RUN
    TRACKING_RUN.log_failed(reason=reason, message=message)


log_failed.__doc__ = Run.log_failed.__doc__


def log_artifact_ref(
    path: str,
    kind: V1ArtifactKind,
    name: str = None,
    hash: str = None,
    content=None,
    summary: Dict = None,
    is_input: bool = False,
    rel_path: str = None,
    skip_hash_calculation: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.log_artifact_ref(
        path=path,
        kind=kind,
        name=name,
        hash=hash,
        content=content,
        summary=summary,
        is_input=is_input,
        rel_path=rel_path,
        skip_hash_calculation=skip_hash_calculation,
    )


log_artifact_ref.__doc__ = Run.log_artifact_ref.__doc__


def log_tensorboard_ref(
    path: str,
    name: str = "tensorboard",
    is_input: bool = False,
    rel_path: str = None,
):
    global TRACKING_RUN
    TRACKING_RUN.log_tensorboard_ref(
        path=path,
        name=name,
        is_input=is_input,
        rel_path=rel_path,
    )


log_tensorboard_ref.__doc__ = Run.log_tensorboard_ref.__doc__


def log_model_ref(
    path: str,
    name: str = None,
    framework: str = None,
    summary: Dict = None,
    is_input: bool = False,
    rel_path: str = None,
    skip_hash_calculation: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.log_model_ref(
        path=path,
        name=name,
        framework=framework,
        summary=summary,
        is_input=is_input,
        rel_path=rel_path,
        skip_hash_calculation=skip_hash_calculation,
    )


log_model_ref.__doc__ = Run.log_model_ref.__doc__


def log_code_ref(code_ref: Dict = None, is_input: bool = True):
    global TRACKING_RUN
    TRACKING_RUN.log_code_ref(code_ref=code_ref, is_input=is_input)


log_code_ref.__doc__ = Run.log_code_ref.__doc__


def log_data_ref(
    name: str,
    hash: str = None,
    path: str = None,
    content=None,
    summary: Dict = None,
    is_input: bool = True,
    skip_hash_calculation: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.log_data_ref(
        name=name,
        content=content,
        hash=hash,
        path=path,
        summary=summary,
        is_input=is_input,
        skip_hash_calculation=skip_hash_calculation,
    )


log_data_ref.__doc__ = Run.log_data_ref.__doc__


def log_file_ref(
    path: str,
    name: str = None,
    hash: str = None,
    content=None,
    summary: Dict = None,
    is_input: bool = False,
    rel_path: str = None,
    skip_hash_calculation: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.log_file_ref(
        path=path,
        name=name,
        hash=hash,
        content=content,
        summary=summary,
        is_input=is_input,
        rel_path=rel_path,
        skip_hash_calculation=skip_hash_calculation,
    )


log_file_ref.__doc__ = Run.log_file_ref.__doc__


def log_dir_ref(
    path: str,
    name: str = None,
    hash: str = None,
    summary: Dict = None,
    is_input: bool = False,
    rel_path: str = None,
    skip_hash_calculation: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.log_dir_ref(
        path=path,
        name=name,
        hash=hash,
        summary=summary,
        is_input=is_input,
        rel_path=rel_path,
        skip_hash_calculation=skip_hash_calculation,
    )


log_dir_ref.__doc__ = Run.log_dir_ref.__doc__


def log_artifact_lineage(body: List[V1RunArtifact]):
    global TRACKING_RUN
    TRACKING_RUN.log_artifact_lineage(body)


log_artifact_lineage.__doc__ = Run.log_artifact_lineage.__doc__


def log_env(rel_path: str = None, content: Dict = None):
    global TRACKING_RUN
    return TRACKING_RUN.log_env(rel_path=rel_path, content=content)


log_env.__doc__ = Run.log_env.__doc__


def sync_events_summaries():
    global TRACKING_RUN
    TRACKING_RUN.sync_events_summaries()


sync_events_summaries.__doc__ = Run.sync_events_summaries.__doc__


def sync_system_events_summaries():
    global TRACKING_RUN
    TRACKING_RUN.sync_system_events_summaries()


sync_system_events_summaries.__doc__ = Run.sync_system_events_summaries.__doc__


def pull_remote_run(
    path: str = None,
    download_artifacts: bool = True,
):
    global TRACKING_RUN
    return TRACKING_RUN.pull_remote_run(
        path=path,
        download_artifacts=download_artifacts,
    )


pull_remote_run.__doc__ = Run.pull_remote_run.__doc__


def push_offline_run(
    path: str,
    upload_artifacts: bool = True,
    clean: bool = False,
):
    global TRACKING_RUN
    TRACKING_RUN.push_offline_run(
        path=path,
        upload_artifacts=upload_artifacts,
        clean=clean,
    )


push_offline_run.__doc__ = Run.push_offline_run.__doc__


def promote_to_model_version(
    version: str,
    description: str = None,
    tags: Union[str, List[str]] = None,
    content: Union[str, Dict] = None,
    connection: str = None,
    artifacts: List[str] = None,
    force: bool = False,
):
    global TRACKING_RUN
    return TRACKING_RUN.promote_to_model_version(
        version=version,
        description=description,
        tags=tags,
        content=content,
        connection=connection,
        artifacts=artifacts,
        force=force,
    )


promote_to_model_version.__doc__ = Run.promote_to_model_version.__doc__


def promote_to_artifact_version(
    version: str,
    description: str = None,
    tags: Union[str, List[str]] = None,
    content: Union[str, Dict] = None,
    connection: str = None,
    artifacts: List[str] = None,
    force: bool = False,
):
    global TRACKING_RUN
    return TRACKING_RUN.promote_to_artifact_version(
        version=version,
        description=description,
        tags=tags,
        content=content,
        connection=connection,
        artifacts=artifacts,
        force=force,
    )


promote_to_artifact_version.__doc__ = Run.promote_to_artifact_version.__doc__


__all__ = [
    "V1ArtifactKind",
    "V1RunArtifact",
    "Run",
    "init",
    "get_or_create_run",
    "get_artifacts_path",
    "get_outputs_path",
    "get_tensorboard_path",
    "set_artifacts_path",
    "set_run_event_logger",
    "set_run_resource_logger",
    "set_run_process_sidecar",
    "log_metric",
    "log_metrics",
    "log_roc_auc_curve",
    "log_sklearn_roc_auc_curve",
    "log_pr_curve",
    "log_sklearn_pr_curve",
    "log_curve",
    "log_confusion_matrix",
    "log_image",
    "log_image_with_boxes",
    "log_mpl_image",
    "log_video",
    "log_audio",
    "log_text",
    "log_html",
    "log_np_histogram",
    "log_histogram",
    "log_model",
    "log_artifact",
    "log_dataframe",
    "log_plotly_chart",
    "log_bokeh_chart",
    "log_altair_chart",
    "log_mpl_plotly_chart",
    "get_log_level",
    "set_description",
    "set_name",
    "end",
    "log_status",
    "log_inputs",
    "log_outputs",
    "log_tags",
    "log_meta",
    "log_progress",
    "log_succeeded",
    "log_stopped",
    "log_failed",
    "log_artifact_ref",
    "log_tensorboard_ref",
    "log_model_ref",
    "log_code_ref",
    "log_data_ref",
    "log_file_ref",
    "log_dir_ref",
    "log_artifact_lineage",
    "log_env",
    "sync_events_summaries",
    "sync_system_events_summaries",
    "pull_remote_run",
    "push_offline_run",
    "promote_to_model_version",
    "promote_to_artifact_version",
]
